@@ -59,8 +59,8 @@ static void constructor(test_batch_runner *runner) {
              "default is list type is bullet");
       INT_EQ(runner, cmark_node_get_list_delim(node), CMARK_NO_DELIM,
              "default is list delim is NO_DELIM");
-      INT_EQ(runner, cmark_node_get_list_start(node), 1,
-             "default is list start is 1");
+      INT_EQ(runner, cmark_node_get_list_start(node), 0,
+             "default is list start is 0");
       INT_EQ(runner, cmark_node_get_list_tight(node), 0,
              "default is list is loose");
       break;
@@ -732,6 +732,14 @@ static void utf8(test_batch_runner *runner) {
       string_with_null, sizeof(string_with_null) - 1, CMARK_OPT_DEFAULT);
   STR_EQ(runner, html, "<p>((((" UTF8_REPL "))))</p>\n", "utf8 with U+0000");
   free(html);
+
+  // Test NUL followed by newline
+  static const char string_with_nul_lf[] = "```\n\0\n```\n";
+  html = cmark_markdown_to_html(
+      string_with_nul_lf, sizeof(string_with_nul_lf) - 1, CMARK_OPT_DEFAULT);
+  STR_EQ(runner, html, "<pre><code>\xef\xbf\xbd\n</code></pre>\n",
+         "utf8 with \\0\\n");
+  free(html);
 }
 
 static void test_char(test_batch_runner *runner, int valid, const char *utf8,
